#!/usr/bin/perl 
# Graph attack trees represented in XML through Graphviz
# inspired by GraphViz::XML http://search.cpan.org/search?dist=GraphViz
# This script is free but comes with no warranty.
# Razvan Peteanu, http://members.rogers.com/razvan.peteanu

use strict;
use vars qw($VERSION);
use Carp;
use GraphViz;
use XML::Twig;


# Constants
my $DIRECTION = 1; # 0 means top to bottom, 1 left to right

my $filename = $ARGV[0];
my $xml = "";
open(FILE,$filename);
while (<FILE>){
    chomp;
    print;
    $xml.=$_;
}
close(FILE);
print "1\n";
my $t = XML::Twig->new();
$t->parse($xml);
print "2\n";
my $graph = GraphViz->new(rankdir  => $DIRECTION);
print "3\n";
parse_node($graph, $t->root, "none");
print "4\n";
$graph->as_png("$filename.png");
print "5\n";
$graph->as_text("$filename.dot");
print "6\n";

sub parse_node {
	my($g, $root, $arrowType) = @_;

	my $label = $root->gi;
	my $colour = 'blue';
	my $shape = 'box';
	my $nodename = $root->gi;

	if ($nodename eq "header") {
		# ignore the header for the graph generation
		return;
	} elsif ($nodename eq "body" || $nodename eq "attack-tree") {
		# do not display elements that don't have to appear in the graph but have their childen processed
		foreach my $child ($root->children) {
			parse_node($g, $child, $arrowType);
		}
		return;
	}				                

	if ($root->is_pcdata) {
		$label = $root->text;
		$label =~ s|^\s+||;
		$label =~ s|\s+$||;
		$colour = 'black';
		$shape = "plaintext";
	} else {
		$shape = "ellipse";
		if ($nodename eq "goal") {
			$label = $root->att("name");
			# colourize contour
			if ($root->att("prevented") eq "y") {
				$colour = 'green';
			} elsif ($root->att("prevented") eq "n") {
				$colour = 'red';
			} else {
				$colour = "black";
			}
			
			#select arrow head depending on the logical operation
			if ($root->att("composed-with") eq "OR") {
				$arrowType = "odot";	# empty circle
			} elsif ($root->att("composed-with") eq "AND") {
				$arrowType = "dot";	# filled circle 
			}
		}
	}

	#add to the graphviz document
	$g->add_node($root, label => $label, color => $colour, shape => $shape, fontsize => 10);
	foreach my $child ($root->children) {
		$g->add_edge($root => $child, dir => 'back', arrowtail => $arrowType, arrowsize => 2);
		parse_node($g, $child, $arrowType);
	}
	
  }


