package Tk::MarkupTools;


use strict;
use vars( qw/$VERSION $top1 $top2 $top3 $BTREE %b/ );
$VERSION = '2.00';

use Tk qw ( 800.022 Ev Exists );
use XML::Parser;
use DB_File;

use AutoLoader;

use base qw( Tk::TextTools );
Construct Tk::Widget 'MarkupTools';
use FindBin qw( $Bin );
use locale;


sub ClassInit {
    
    my ( $class, $parent ) = @_;
    $class->SUPER::ClassInit( $parent );
    
    $parent->bind( $class, '<Control-l>', 'GotoLineNumberPopUp' );
    $parent->bind( $class, '<Control-m>', 'MarkupPopUp' );
    $parent->bind( $class, '<Control-t>', 'WriteTagPopUp' );
 
    $parent->bind( $class, '<F5>',        'IndentSelectedLines' );
    $parent->bind( $class, '<F7>',        'showIndex' );
    $parent->bind( $class, '<Control-p>', 'showPreferences' );
    
    return( $class );
}


sub InitObject {
    
    my ( $self ) = @_;
    $self->SUPER::InitObject;
    
    $self->getEditPreferences();
    $self->resetXMLIndicators();
    return( $self );
}


sub setHomeEnvironement {

    my ( $self, $home ) = @_;
    return( $self->{HOME}=$home );
}


sub getEditPreferences {
    
    my ( $self ) = @_;
    
    if ( $self->{HOME} && open( CONF, "<".$self->{HOME}."/.edxmlrc" )) {}
    elsif ( open( CONF, "<".$Bin."/ledxml/.edxmlrc" )) {}
    else { die "Cannot read the configuration file! $!\n" ; }
    
    while ( <CONF> ) {
	if ( /^([A-Z_]+)=(.+)$/ ) {
	    $self->{$1}=$2;
	}
	else { die "The configuration file is dammaged! $!\n"; }
    }
    close( CONF );
    return( $self );
}


sub resetXMLIndicators {

    my ( $self ) = @_;

    $self->{LAST_INDEX} = '1.0'; $self->{ELEMENT_STACK} = ();
    $self->{ELEMENT_LEVEL} = 0; $self->{CONTROL_LEVEL} = 0;

    $self->{SEEN} = (); $self->{HAS_DATA} = 0;
    $self->{HAS_DATA_STACK} = (); $self->{HAS_ELEMENT} = 0;
    $self->{HAS_ELEMENT_STACK} = ();

    return( $self );
}


sub IndentSelectedLines {
    
    my ( $self ) = @_;

    if ( $self->getSafeMode ) {
	
	$self->cancel_current_gui_callback_and_restart_from_beginning;
	my ( $line, $column ) = split( /\./, $self->index( 'insert' ));
	if ( $self->get( "$line.0 linestart", "$line.0 lineend" ) &&
	     $self->compare( $self->{LAST_INDEX}, "!=", "$line.0 lineend" )) {
	    
	    $self->resetXMLIndicators();
	    $self->parseEdits( $self->get( '1.0', "$line.0 lineend" ));
	    
	    if ( $self->{DATA_MODE} ) {
		
		$self->markSet( 'insert', "$line.0 linestart" );
		for ( 1 .. (( $self->{ELEMENT_LEVEL} - $self->{CONTROL_LEVEL})
			    * $self->{DATA_INDENT} )) {
		    $self->insert( 'insert', $self->{INDENT_STRING} );
		}
		while ( $self->get( 'insert', 'insert + 1 char' ) =~ /\s/ ) {
		    $self->delete( 'insert', 'insert + 1 char' );
		}
	    }
	}
	$self->markSet( 'insert', "$line.0 lineend" );
	$self->{LAST_INDEX} = $self->index( 'insert' );
    }

    return( $self );
}


sub parseEdits {
    
    my ( $self, $data ) = @_;

    my ( $parser ) = new XML::Parser();

    $self->Busy();
    $self->tellStatusbar( "Parsing edits..." );

    $parser->setHandlers(

    XMLDecl => sub { ${$self->{SEEN}}{ANYTHING} = 1;
		     ${$self->{SEEN}}{XMLDECL}  = 1; },
    Doctype => sub { ${$self->{SEEN}}{ANYTHING} = 1;
		     ${$self->{SEEN}}{DOCTYPE}  = $_[1]; },
    Comment => sub { ${$self->{SEEN}}{ANYTHING} = 1; },
    Start   => sub { ${$self->{SEEN}}{ANYTHING} = 1;
		     ${$self->{SEEN}}{ELEMENT}  = 1;
		     $self->{ELEMENT_LEVEL}++;
                     $self->{CONTROL_LEVEL} = 1;
                     push @{$self->{ELEMENT_STACK}}, $_[1];
                     if ( $self->{DATA_MODE} ) {
	             $self->{HAS_ELEMENT} = 1;
	             push @{$self->{HAS_DATA_STACK}}, $self->{HAS_DATA};
	             $self->{HAS_DATA} = 0;
	             push @{$self->{HAS_ELEMENT_STACK}}, $self->{HAS_ELEMENT};
		     $self->{HAS_ELEMENT} = 0; }},
    Char    => sub { if ( $self->{DATA_MODE} && $_[1] =~ /\S/ ) {
	             $self->{HAS_DATA} = 1; }},
    End     => sub { pop( @{$self->{ELEMENT_STACK}} );
		     $self->{ELEMENT_LEVEL}--;
		     $self->{CONTROL_LEVEL} = 0;
		     if ( $self->{DATA_MODE} ) {
		     $self->{HAS_DATA} = pop @{$self->{HAS_DATA_STACK}};
		     $self->{HAS_ELEMENT}=pop@{$self->{HAS_ELEMENT_STACK}};}});
    
    eval { $parser->parse( $data ); };
    $self->tellStatusbar( "Edits were parsed." );
    $self->Unbusy();
  
    return( $self );
}


sub MenuLabels {

    my ( $self ) = @_;
    return( qw[~File ~Edit ~Search ~Markup ~Tools ~View] );
}


sub FileTypes {

    my ( $self ) = @_;
    return( [["Xml Files",'.xml'],["Text Files",'.txt'],["All Files",'*']] );
}


sub MarkupMenuItems {
    
    my ( $self ) = @_;
    
    return( [['command'=>'Display Tag ~Area',-command=>[$self,'displayArea']],
	     ['command'=>'~Delete Tag Area',-command=>[$self,'deleteArea']],
	     ['command'=>'~Remove Markups',-command=>[$self,'removeMarkups']],
	     '-',
	     ['command'=>'~XML Declaration',-command=> [$self,'XMLDeclPopUp']],
	     ['command'=>'Document ~Type',-command=> [$self,'DoctypePopUp']],
	     ['command'=>'Insert ~Comment',-command=> [$self,'CommentPopUp']],
	     '-',
	     ['command'=>'~Write Tag',-command=> [$self, 'WriteTagPopUp'],
	      -accelerator=> 'Ctrl+T'],
	     ['command'=>'~Insert Text',-command=> [$self,'InsertTextPopUp']],
	     '-',
	     ['command'=>'~Markup Selections',-command=> [$self,'MarkupPopUp'],
	      -accelerator => 'Ctrl+M']] );
}


sub ToolsMenuItems {

    my ( $self ) = @_;
    
    return( [@{$self->SUPER::ToolsMenuItems},
	     ['command'    => 'Show ~Index',
	      -command     => [$self,'showIndex'],
	      -accelerator => 'F7']] );
}


sub ViewMenuItems {

    my ( $self ) = @_;
    my ( $var );

    tie $var,'Tk::Configure', $self, '-wrap';
    
    return( [['command'     => 'Goto ~Line...',
	      -command      => [$self, 'GotoLineNumberPopUp'],
	      -accelerator  => 'Ctrl+L'],
	     ['cascade'     => '~Wrap', -tearoff => 0,
	      -menuitems    => [[radiobutton    => 'Word',
				 -variable      => \$var,
				 -value         => 'word'],
				[radiobutton    => 'Character',
				 -variable      => \$var,
				 -value         => 'char'],
				[radiobutton    => 'None',
				 -variable      => \$var,
				 -value         => 'none']]],
	     '-',
	     ['command'     => "~Preferences",
	      -command      => [$self, 'showPreferences'],
	      -accelerator  => 'Ctrl+P'],
	     ]);
}


sub FindNextTag {
    
    my ( $self, $direction, $mode, $case, $pattern ) = @_;
    my ( $start_index, $end_index, $return );

    eval {
	if ( $direction eq '-forward' ) {
	    $self->markSet( 'insert', 'sel.last' );
	    $self->markSet( 'current', 'sel.last' );
	}
	else {
	    $self->markSet( 'insert', 'sel.first' );
	    $self->markSet( 'current', 'sel.first' );
	}
    };

    my ( $saved_index ) = $self->index( 'insert' );
    $self->unselectAll;
    my ( $match_length ) = 0;

    if ( $case eq '-nocase' ) {

	unless (( $start_index = $self->search
		  ( $direction,
		    '-regexp',
		    $case,
		    "--",
		    "<$pattern(?: [^>]+)*>",
		    'insert',
		    $direction eq '-forward'? 'end': '1.0' )) &&

		( $end_index = $self->search
		  ( '-forward',
		    $mode,
		    $case,
		    -count=>\$match_length,
		    "--",
		    "</$pattern>",
		    $start_index,
		    'end' ))) {
	    return( 0 );
	}
    }

    else {

	unless (( $start_index = $self->search
		  ( $direction,
		    '-regexp',
		    "--",
		    "<$pattern(?: [^>]+)*>",
		    'insert',
		    $direction eq '-forward'? 'end': '1.0' )) &&

		( $end_index = $self->search
		  ( '-forward',
		    $mode,
		    -count=>\$match_length,
		    "--",
		    "</$pattern>",
		    $start_index,
		    'end' ))) {
	    return( 0 );
	}
    }

    my ( $line, $column ) = split( /\./, $end_index );
    $column = $column + $match_length;
    $end_index = $line.'.'.$column;
    
    $self->tagAdd( 'sel', $start_index, $end_index );
    $self->see( $start_index );
    
    if ( $direction eq '-forward' ) {
	
	$self->markSet( 'insert', $end_index );
	$self->markSet( 'current', $end_index );
    }

    else {
	
	$self->markSet( 'insert', $start_index );
	$self->markSet( 'current', $start_index );
    }
	
    my ( $compared_index ) = $self->index( 'insert' );
    if ( $compared_index eq $saved_index ) {$return = 0;}
    else {$return = 1;}

    return( $return );
}


sub FindAllTags {

    my ( $self, $mode, $case, $pattern ) = @_;

    my ( $end_index ) = '1.0';
    $self->unselectAll;

    while( defined( $end_index )) {

	my ( $start_index );
	my ( $match_length ) = 0;

	if ( $case eq '-nocase' ) {

	    unless (( $start_index = $self->search
		      ( '-regexp',
			$case,
			"--",
			"<$pattern(?: [^>]+)*>",
			$end_index,
			'end' )) &&
		    
		    ( $end_index = $self->search
		      ( $mode,
			$case,
			-count=>\$match_length,
			"--",
			"</$pattern>",
			$start_index,
			'end' ))) {
		last();
	    }
	}
	
	else {
	    
	    unless (( $start_index = $self->search
		      ( '-regexp',
			"--",
			"<$pattern(?: [^>]+)*>",
			$end_index,
			'end' )) &&

		    ( $end_index = $self->search
		      ( $mode,
			-count=>\$match_length,
			"--",
			"</$pattern>",
			$start_index,
			'end' ))) {
		last();
	    }
	}
	
	my ( $line, $colon ) = split( /\./, $end_index );
	$colon = $colon + $match_length;
	$end_index = $line.'.'.$colon;
	
	$self->tagAdd( 'sel', $start_index, $end_index );
    }
    return( $self );
}


sub displayArea {

    my ( $self ) = @_;
    return( $self->markuppopup( "Display Tag Area" ));
}


sub deleteArea {

    my ( $self )= @_;
    return( $self->markuppopup( "Delete Tag Area" ));
}


sub removeMarkups {
 
    my ( $self ) = @_;

    my ( $message ) = "All the standard markups will be removed ! Proceed ?" ;
    my ( $answer ) = $self->messageBox( -icon    => 'warning',
				 -title   => 'Warning',
				 -type    => 'OKCancel',
				 -message =>  $message );
    if ( $answer eq 'cancel' ) { return( 0 );}
    return( $self->FindAndReplaceAll( \&Tk::TextTools::FindNext,
				      \&Tk::TextTools::ReplaceSelectionsWith,
				      '-regexp', '-nocase', "<[^>]+>", '' ));
}


sub XMLDeclPopUp {

    my ( $self )= @_;

    if (( $self->{SAFE_MODE} ) &&
	( $self->{LAST_INDEX} != $self->index( 'insert' ))) {
	return( 0 );
    }
    return( $self->markuppopup( "Insert Declaration" ));
}


sub XMLDecl {
    
    my ( $self, $version, $encoding, $standalone ) = @_;

    ${$self->{SEEN}}{ANYTHING} = 1;
    ${$self->{SEEN}}{XMLDECL}  = 1;

    if ( $standalone && $standalone ne 'no' ) {
	$standalone = 'yes';
    }
    $encoding = "ISO-8859-1" unless $encoding;
    $version  = "1.0" unless $version;
    
    my ( $xml ) .= "<?xml version=\"$version\"";
    if ( $encoding ) {
	$xml .= " encoding=\"$encoding\"";
    }
    if ( $standalone ) {
	$xml .= " standalone=\"$standalone\"";
    }
    $xml .= "?>\n";
    return( $self->insert( 'insert', $xml ));
}


sub SafeXMLDecl {
    
    my ( $self, $version, $encoding, $standalone ) = @_;
    
    if ( ${$self->{SEEN}}{ANYTHING} ) {
	$self->BackTrace
	( "The XML declaration is not the first thing in the document" );
    }
    return( $self );
}


sub DoctypePopUp {

    my ( $self )= @_;

    if (( $self->{SAFE_MODE} ) &&
	( $self->{LAST_INDEX} != $self->index( 'insert' ))) {
	return( 0 );
    }
    return( $self->markuppopup( "Insert Doctype" ));
}


sub Doctype {

    my ( $self, $name, $sysid, $pubid ) = @_;

    ${$self->{SEEN}}{ANYTHING} = 1;
    ${$self->{SEEN}}{DOCTYPE}  = $name;

    my ( $xml ) .= "<!DOCTYPE $name";
    if ( $pubid ) {
	$xml .= " PUBLIC \"$pubid\" \"$sysid\"";
    }
    elsif ( $sysid ) {
	$xml .= " SYSTEM \"$sysid\"";
    }
    $xml .= ">\n";
    return( $self->insert( 'insert', $xml ));
}


sub SafeDoctype {

    my ( $self, $name, $sysid, $pubid ) = @_;
    
    if ( ${$self->{SEEN}}{DOCTYPE} ) {
	$self->BackTrace( "Attempt to insert second DOCTYPE declaration" );
    }
    elsif ( ${$self->{SEEN}}{ELEMENT} ) {
	$self->BackTrace
        ( "The DOCTYPE declaration must come before the first start tag" );
    }
    return( $self );
}


sub CommentPopUp {

    my ( $self )= @_;

    if (( $self->{SAFE_MODE} ) &&
	( $self->{LAST_INDEX} != $self->index( 'insert' ))) {
	return( 0 );
    }
    return( $self->markuppopup( "Insert Comment" ));
}


sub Comment {
    
    my ( $self, $data ) = @_;

    ${$self->{SEEN}}{ANYTHING} = 1;

    my ( $xml ) .= "<!--". $data ."-->";
    if ( $self->{ELEMENT_LEVEL} == 0 ) {
	$xml .= "\n";
    }
    return( $self->insert( 'insert', $xml ));
}


sub SafeComment {
    
    my ( $self, $data ) = @_;
    
    if ( $data =~ /--/ ) {
	my ( $answer );

	 $answer = $self->messageBox( -icon => 'warning',
				 -title => 'Warning',
				 -type  => 'OkCancel',
				 -default => 'Ok',
				 -message =>
			  "Interoperability problem: \"--\" in comment text" );
	if ( $answer eq 'Cancel' ) { return( 0 ); }
    }

    if ( $data =~ /-->/ ) {
	$self->BackTrace( "Comment may not contain '-->'" );
    }
    return( $self );
}


sub InsertTextPopUp {

    my ( $self )= @_;

    if (( $self->{SAFE_MODE} ) &&
	( $self->{LAST_INDEX} != $self->index( 'insert' ))) {
	return( 0 );
    }
    return( $self->markuppopup( "Insert Text" ));
}


sub Text {
    
    my ( $self, $data ) = @_;
    
    my ( $xml ) = $self->_escapeLiteral( $data );
    $self->{HAS_DATA} = 1;
    return( $self->insert( 'insert', $xml ));
}


sub SafeText {
    
    my ( $self, $data ) = @_;
    if ( $self->{ELEMENT_LEVEL} < 1 ) {
	
	$self->BackTrace
	    ( "Attempt to insert characters outside of document element" );
    }
    elsif ( $self->{DATA_MODE} && $self->{HAS_ELEMENT} ) {
	$self->BackTrace
	    ( "Mixed content not allowed in data mode: characters" );
    }
    return( $self );
}


sub _escapeLiteral {

    my ( $self, $data ) = @_;

    if ( $data =~ /[\&\<\>\"]/ ) {

	$data =~ s/\&/\&amp\;/g;
	$data =~ s/\</\&lt\;/g;
	$data =~ s/\>/\&gt\;/g;
	$data =~ s/\"/\&quot\;/g;
    }
    return( $data );
}


sub WriteTagPopUp {

    my ( $self )= @_;
    
    if (( $self->{SAFE_MODE} ) &&
	( $self->{LAST_INDEX} != $self->index( 'insert' ))) {
	return( 0 );
    }
    return( $self->markuppopup( "Write Tag" ));
}


sub _checkAttributes {
    
    my ( $self, @atts ) = @_;
    my ( $i, %sort );
    
    for ( $i = 0; $i <= $#atts -1; $i+=2 ) {
	if ( defined( $atts[$i] ) and length( $atts[$i] )) {
	    if ( exists( $sort{$atts[$i]} )) {
		$self->BackTrace( "Two attributes named \"$atts[$i]\"" );
	    }
	    else { $sort{$atts[$i]} = 1; }
	}
    }
    return( $self );
}


sub _showAttributes {

    my ( $self, $xml, @atts ) = @_;
    my ( $i );
    
    for ( $i = 0; $i <= $#atts -1; $i+=2 ) {
	if ( defined( $atts[$i] ) and length( $atts[$i] )) {
	    
	    $xml .= " $atts[$i]";
	    if ( defined( $atts[$i+1] ) and length( $atts[$i+1])) {
		
		$atts[$i+1] = $self->_escapeLiteral( $atts[$i+1] );
		$xml .= "=\"$atts[$i+1]\"";
	    }
	}
    }
    return( $xml );
}


sub StartTag {
    
    my ( $self, $name, @atts ) = @_;
    my ( $xml );
 
    ${$self->{SEEN}}{ANYTHING} = 1;
    ${$self->{SEEN}}{ELEMENT}  = 1;

    if ( $self->{DATA_MODE} ) {
	$xml .= "\n";
	for ( 1 .. ( $self->{ELEMENT_LEVEL} * $self->{DATA_INDENT} )) {
	    $xml .= $self->{INDENT_STRING};
	}
    }
    
    $self->{ELEMENT_LEVEL}++;
    push @{$self->{ELEMENT_STACK}}, $name;
    
    $xml .= "<$name";
    $xml = $self->_showAttributes( $xml, @atts );
    $xml .= ">";
    
    if ( $self->{DATA_MODE} ) {
	$self->{HAS_ELEMENT} = 1;
	push @{$self->{HAS_DATA_STACK}}, $self->{HAS_DATA};
	$self->{HAS_DATA} = 0;
	push @{$self->{HAS_ELEMENT_STACK}}, $self->{HAS_ELEMENT};
	$self->{HAS_ELEMENT} = 0;
    }
    return( $self->insert( 'insert', $xml ));
}


sub SafeStartTag {
    
    my ( $self, $name, @atts ) = @_;
    
    $self->_checkAttributes( @atts );
    
    if (( ${$self->{SEEN}}{ELEMENT} ) && ( $self->{ELEMENT_LEVEL} == 0 )) {
	$self->BackTrace
	    ( "Attempt to insert start tag after close of document element" );
    }

    if (( $self->{ELEMENT_LEVEL} == 0 ) && ( ${$self->{SEEN}}{DOCTYPE} ) &&
	( $name ne ${$self->{SEEN}}{DOCTYPE} )) {
	$self->BackTrace( "Document element is \"$name\", but DOCTYPE is \""
			  . ${$self->{SEEN}}{DOCTYPE}
			  . "\"" );
    }
    elsif ( $self->{DATA_MODE} && $self->{HAS_DATA} ) {
	$self->BackTrace
	    ( "Mixed content not allowed in data mode: element $name" );
    }
    return( $self );
}


sub EmptyTag {
    
    my ( $self, $name, @atts ) = @_;
    my ( $xml );
    
    ${$self->{SEEN}}{ANYTHING} = 1;
    ${$self->{SEEN}}{ELEMENT}  = 1;
    
    if ( $self->{DATA_MODE} ) {
	$xml .= "\n";
	for ( 1 .. ( $self->{ELEMENT_LEVEL} * $self->{DATA_INDENT} )) {
	    $xml .= $self->{INDENT_STRING};
	}
    }

    $xml .= "<$name";
    $xml = $self->_showAttributes( $xml, @atts );
    $xml .= "/>";
    
    if ( $self->{DATA_MODE} ) {
	$self->{HAS_ELEMENT} = 1;
    }
    return( $self->insert( 'insert', $xml ));
}


sub SafeEmptyTag {
    
    my ( $self, $name, @atts ) = @_;
    
    $self->_checkAttributes( @atts );
    if ( ${$self->{SEEN}}{ELEMENT} && $self->{ELEMENT_LEVEL} == 0 ) {
	$self->BackTrace
	    ( "Attempt to insert empty tag after close of document element" );
    }
    elsif (( $self->{ELEMENT_LEVEL} == 0 ) &&
	   ( ${$self->{SEEN}}{DOCTYPE} ) &&
	   ( $name ne ${$self->{SEEN}}{DOCTYPE} )) {
	$self->BackTrace( "Document element is \"$name\", but DOCTYPE is \""
			  .  ${$self->{SEEN}}{DOCTYPE}
			  . "\"");
    }
    elsif (( $self->{DATA_MODE} ) && ( $self->{HAS_DATA} )) {
	$self->BackTrace
	    ( "Mixed content not allowed in data mode: element $name" );
    }
    return( $self );
}



sub EndTag {
    
    my ( $self, $name ) = @_;
    my ( $xml );
    
    my ( $currentName ) = pop @{$self->{ELEMENT_STACK}};
    $name = $currentName unless ( $name );
    
    $self->{ELEMENT_LEVEL}--;
    if ( $self->{DATA_MODE} && $self->{HAS_ELEMENT} ) {
	$xml .= "\n";
	for ( 1 .. ( $self->{ELEMENT_LEVEL} * $self->{DATA_INDENT} )) {
	    $xml .= $self->{INDENT_STRING};
	}
    }
    
    $xml .= "</$name>";
    if ( $self->{DATA_MODE} ) {
	$self->{HAS_DATA} = pop @{$self->{HAS_DATA_STACK}};
	$self->{HAS_ELEMENT} = pop @{$self->{HAS_ELEMENT_STACK}};
    }
    return( $self->insert( 'insert', $xml ));
}


sub SafeEndTag {
    
    my ( $self, $name ) = @_;    
    my ( $oldName ) = ${$self->{ELEMENT_STACK}}[$#{$self->{ELEMENT_STACK}}];

    if ( $self->{ELEMENT_LEVEL} <= 0 ) {
	$self->BackTrace
	    ( "End tag \"$name\" does not close any open element" );
    }
    elsif ( $name && $name ne $oldName ) {
	$self->BackTrace
	    ( "Attempt to end element \"$oldName\" with \"$name\" tag" );
    }
    else {
	return( $self );
    }
}


sub MarkupPopUp {
    my ( $self )= @_;

    if (! $self->getSafeMode ) {
	return( $self->markuppopup( "Markup Selections" ));
    }
    return( 0 );
}


sub markupSelections {
    
    my ( $self, $tag, @atts ) = @_;
    my ( $data );
    
    eval { $data = $self->SelectionGet( -selection=>"PRIMARY" );};
    unless ( $@ ) {
	if ( defined( $data ) and length( $data )) {
	    my ( $line, $column, $saved_index );
	    
	    $self->markSet( 'insert', 'sel.first' );
	    $saved_index = $self->index( 'insert' );
	    $self->unselectAll;

	    $self->StartTag( $tag, @atts );
	    ( $line, $column ) = split( /\./, $self->index( 'insert' ));
	    $self->markSet( 'insert', $line.".".( $column + length( $data )));
	    $self->EndTag( $tag );

	    $self->tagAdd( 'sel', $saved_index, 'insert' );
	}
    }
    return( $self );
}


sub markuppopup {
    
    my ( $self, $action )= @_;
    if (! Exists( $top1 )) {

	my ( $label1, $pattern1, $label2, $pattern2, $label3, $pattern3,
	     $pattern4, $pattern5, $radio, $checked1, $checked2, $checked3,
	     @ranges, $selection, $first, $last,
	     $frame6, $attribute1, $attribute2, $attribute3,
	     $value1, $value2, $value3, @atts,
	     $button_ok,$button_next, $button_all, $button_cancel,
	     $direction, $mode, $case );
    
	$self->tellStatusbar( "Opening assisted session..." );
	$top1 = $self->Toplevel( -title => $action );
    
	my ( $frame1 ) = $top1->Frame(); 
	$frame1->Label( -textvariable=> \$label1 )->grid
	    ( -row => 0, -column=> 0 );
	my ( $entry1 ) = $frame1->Entry( -textvariable => \$pattern1,
					 -width => 35,
					 -selectforeground => 'white',
					 -selectbackground => "#000070669645")
	    ->grid( -row => 0, -column => 1, -ipady => 5 );
	$entry1->bind( '<Button-3>', sub {
	
	    eval {
		$selection = $self->SelectionGet( -selection => "PRIMARY" );};
	    unless( $@ ) {
		if ( defined( $selection ) and length( $selection )) {
		    $entry1->insert( 'end', $self->makeRegex
				     ( split( /\s+/, $selection )));
		}
	    }
	});

	my ( $frame2 ) = $top1->Frame();
	$frame2->Label( -textvariable => \$label2 )->grid
	    ( -row => 0, -column=>0 );
	$frame2->Entry( -textvariable => \$pattern2, -width => 35,
			-selectforeground => 'white',
		    -selectbackground => "#000070669645" )->grid
			( -row => 0, -column => 1, -ipady => 5 );

	my ($frame3 ) = $top1->Frame();
	$frame3->Label( -textvariable => \$label3 )->grid
	    ( -row => 0, -column=>0, -sticky=> 'nsew' );
	my ( $entry3 ) = $frame3->Entry( -textvariable => \$pattern3,
					 -width => 40,
					 -selectforeground => 'white',
			      -selectbackground => "#000070669645" )->grid
				  ( -row => 1, -column => 0 );
	$frame3->Entry( -textvariable => \$pattern4, -width => 40,
			-selectforeground => 'white',
			-selectbackground => "#000070669645" )->grid
			    ( -row => 2, -column => 0 );
	$frame3->Entry( -textvariable => \$pattern5, -width => 40,
			-selectforeground => 'white',
			-selectbackground => "#000070669645" )->grid
			    ( -row => 3, -column => 0 );

	my ( $frame4 ) = $top1->Frame();
	$frame4->Label( -text => "Text String: " )->grid
	    ( -row => 0, -column=> 0, -sticky => 'nsew', -ipady => 3 );
	my ( $text ) = $frame4->Text( -width => 40, -height => 5,
			   -selectforeground => 'white',
			   -selectbackground => "#000070669645" )->grid
			       ( -row => 1, -column => 0, -ipady => 5 );
    
	my ( $frame5 ) = $top1->Frame();
	$frame5->Radiobutton( -text => "Start Tag", -variable => \$radio,
			      -value => 1, -command => sub {
			      if ( defined($frame6) && (!$frame6->ismapped)) {
				  $frame6->grid( -row => 3, -column => 0,
						 -ipadx =>10,-ipady => 5,
						 -sticky => 'ew' );
			      }
			  })->grid( -row => 0, -column => 0, -padx => 15,
				    -sticky => 'nsew')->invoke();
	$frame5->Radiobutton( -text => "Empty Tag", -variable => \$radio,
			      -value => 2, -command => sub {
			      if ( defined($frame6) && (!$frame6->ismapped)) {
				  $frame6->grid( -row => 3, -column => 0,
						 -ipadx =>10,-ipady => 5,
						 -sticky => 'ew' );
			      }
			  })->grid( -row => 0, -column => 1,
				    -padx=>15, -sticky => 'nsew' );
	$frame5->Radiobutton( -text => "End Tag", -variable => \$radio,
			      -value => 3, -command => sub {
				  if ( $frame6->ismapped ) {
				      $frame6->gridForget();
				  }
			      })->grid( -row => 0, -column => 2,
					-padx => 15, -sticky => 'nsew');

	$frame6 = $top1->Frame();
	$frame6->Label( -text => "Tag Attributes:")->grid
	    ( -row => 0, -column => 0, -sticky =>'nsew' );
	$frame6->Label( -text=>"Tag Values:")->grid
	    ( -row => 0, -column=> 1, -sticky => 'nsew' );
	$frame6->Entry( -textvariable => \$attribute1, -width => 24 )->grid
	    ( -row => 1, -column => 0, -sticky => 'ew' );
	$frame6->Entry( -textvariable => \$value1,-width => 24 )->grid
	    ( -row => 1, -column => 1, -sticky => 'ew');
	$frame6->Entry( -textvariable => \$attribute2, -width=>24 )->grid
	( -row => 2, -column => 0, -sticky =>'ew' );
	$frame6->Entry( -textvariable => \$value2, -width =>24 )->grid
	    ( -row => 2, -column => 1, -sticky => 'ew');
	$frame6->Entry( -textvariable => \$attribute3, -width => 24 )->grid
	    ( -row => 3, -column => 0, -sticky => 'ew' );
	$frame6->Entry( -textvariable => \$value3, -width => 24 )->grid
	    ( -row => 3, -column => 1, -sticky => 'ew' );


	my ( $frame7 ) = $top1->Frame();
	$frame7->Checkbutton( -text => "Search Backwards",
			      -variable => \$checked1 )->grid
				  ( -row => 0, -column => 0, -sticky => 'w' );
	$frame7->Checkbutton( -text => "Regexp Search",
			      -variable => \$checked2 )->grid
				  ( -row => 0, -column => 1, -sticky=>'w' );
	$frame7->Checkbutton( -text => "Ignore Case",
			      -variable => \$checked3 )->grid
				  (-row=>0,-column=>2, -sticky=>'w' );


	my ($frame8 ) = $top1->Frame( -relief => 'groove',
				     -borderwidth => 2 )->grid
	( -row => 4,-column => 0, -ipadx => 10, -ipady => 10,-sticky => 'ew' );
	$button_ok = $frame8->Button( -text => 'OK' )->grid
	    ( -row => 0, -column => 0,-sticky => 'nsew',
	      -padx => $action =~ /^(?:Write|Insert)/ ? 60: 15 );
	$button_next = $frame8->Button( -text => 'Next' )->grid
	    ( -row => 0, -column => 1, -padx => 15, -sticky => 'nsew' );
	$button_all = $frame8->Button( -text => 'All' )->grid
	    ( -row => 0, -column => 2, -padx => 15, -sticky => 'nsew');
	$button_cancel = $frame8->Button( -text => 'Cancel', -command => sub {
	    $self->tellStatusbar( "Assisted session closed." );
	    $self->{LAST_INDEX} = $self->index( 'insert' );
	    return( $top1->destroy() );})->grid
		( -row=>0,-column=>3,-sticky=>'nsew',
		  -padx => $action =~ /^(?:Write|Insert)/ ? 40: 15 );

    
	@ranges = $self->tagRanges( 'sel' );
	if ( @ranges ) {
	
	    my ( $first_line, $first_colon, $last_line, $last_colon );
	
	    $first = shift( @ranges );
	    $last = shift( @ranges );
	    ( $first_line, $first_colon ) = split( /\./, $first );
	    ( $last_line, $last_colon ) = split( /\./, $last );
	    
	    unless( $first_line == $last_line ) { $last = $first. ' lineend'; }
	    $pattern1 =  $self->get( $first , $last );
	    $entry1->selectionTo( 'end' );
	    $entry1->focus();
	}
	else {	
	    eval { $selection = $self->SelectionGet( -selection=>"PRIMARY" );};
	    unless( $@ ) {
		if ( defined( $selection )) {
		    
		    $pattern1 =  $selection ;
		    $entry1->selectionTo( 'end' );
		    $entry1->focus();
		}
	    }
	}
    
	if ( $action eq "Display Tag Area" ) {
	    
	    $label1   = "Tag Name: ";
	    $frame1->grid( -row => 0, -column => 0, -ipadx => 10,
			   -ipady => 5, -sticky => 'ew' ); 
	    $frame7->grid( -row => 1,-column => 0, -ipadx => 10,
			   -ipady => 5, -sticky => 'w' );
	    $button_ok->gridForget();

	    $button_next->configure( -command => sub {
		
		$direction = $checked1? '-backward': '-forward';
		$mode      = $checked2? '-regexp': '-exact';
		$case      = $checked3? '-nocase': '-case';
		
		if ( defined( $pattern1 ) and length( $pattern1 )) {
		    $self->FindNextTag( $direction, $mode, $case, $pattern1 );
		}
	    });
	    $button_all->configure( -command => sub {
		
		$mode      = $checked2? '-regexp': '-exact';
		$case      = $checked3? '-nocase': '-case';
		
		if ( defined( $pattern1 ) and length( $pattern1 )) {    
		    $self->FindAllTags( $mode, $case, $pattern1 );
		}
	    });
	}
	
	elsif ( $action eq "Delete Tag Area" ) {
	    
	    $label1   = "Tag Name: ";
	    $frame1->grid( -row => 0, -column => 0, -ipadx => 10,
			   -ipady => 5, -sticky => 'ew' ); 
	    $frame7->grid( -row => 1,-column => 0, -ipadx => 10,
			   -ipady => 5, -sticky => 'w' );
	    
	    $button_ok->configure( -command => sub {
		
		$self->deleteSelected();
		$button_next->invoke();
	    });
	    $button_next->configure( -command => sub {
		
		$direction = $checked1? '-backward': '-forward';
		$mode      = $checked2? '-regexp': '-exact';
		$case      = $checked3? '-nocase': '-case';
		
		if ( defined( $pattern1 ) and length( $pattern1 )) {
		    $self->FindNextTag( $direction, $mode, $case, $pattern1 );
		}
	    });
	    $button_all->configure( -command => sub {
		
		$mode      = $checked2? '-regexp': '-exact';
		$case      = $checked3? '-nocase': '-case';
		
		if ( defined( $pattern1 ) and length( $pattern1 )) {
		    
		    $self->FindAndReplaceAll( \&FindNextTag,
					\&Tk::TextTools::ReplaceSelectionsWith,
					  $mode, $case, $pattern1, '' );
		}
	    });
	}
	
	elsif ( $action eq "Insert Declaration" ) {
	    
	    $label3   = "Version, Encoding, Standalone:";
	    $frame3->grid( -row => 0, -column => 0, -ipadx => 10,
			   -ipady=> 5, -sticky=> 'ew' );
	    
	    $pattern3 = "1.0";
	    $pattern4 = "ISO-8859-1";
	    $pattern5 = "yes";
	    
	    $entry3->selectionTo( 'end' );
	    $entry3->focus();
	    
	    $button_ok->configure( -command => sub {
		
		if (( $self->{SAFE_MODE} ) &&
		    ( $self->{LAST_INDEX} == $self->index( 'insert' ))) {
		    $self->SafeXMLDecl( $pattern3, $pattern4, $pattern5 );
		}
		$self->XMLDecl( $pattern3, $pattern4, $pattern5 );
		$self->{LAST_INDEX} = $self->index( 'insert' );
		$button_cancel->invoke();
	    });
	    $button_next->gridForget();
	    $button_all->gridForget();
	}
	
	elsif ( $action eq "Insert Doctype" ) {
	    
	    $label3   = "Name, System id, Public id:";
	    $frame3->grid( -row => 0, -column => 0, -ipadx => 10,
			   -ipady=> 5, -sticky=> 'ew' );
	    $pattern3 = "ecollateral";
	    $entry3->selectionTo( 'end' );
	    $entry3->focus();
	    
	    $button_ok->configure( -command => sub {
		
		if ( defined( $pattern3 ) and length( $pattern3 )) {
		    if ( $self->{SAFE_MODE} ) {
			$self->SafeDoctype( $pattern3, $pattern4, $pattern5 );
		    }
		    $self->Doctype( $pattern3, $pattern4, $pattern5 );
		    $self->{LAST_INDEX} = $self->index( 'insert' );
		    $button_cancel->invoke();
		}
	    });
	    
	    $button_next->gridForget();
	    $button_all->gridForget();
	}
	
	elsif ( $action eq "Insert Comment" ) {
	    
	    $frame4->grid( -row => 0, -column => 0, -ipadx => 15,
			   -ipady => 5, -sticky => 'ew' );
	    
	    $button_ok->configure( -command => sub {
		
		my ( $string ) = $text->get( '1.0', 'end' );
		chomp( $string );
		
		if ( defined( $string ) and length( $string )) {
		    if ( $self->{SAFE_MODE} ) {
			$self->SafeComment( $string );
		    }
		    $self->Comment( $string );
		    $self->{LAST_INDEX} = $self->index( 'insert' );
		    $button_cancel->invoke();
		}
	    });
	    $button_next->gridForget();
	    $button_all->gridForget();
	}
	elsif ( $action eq "Insert Text" ) {
	    
	    $frame4->grid( -row => 0, -column => 0, -ipadx => 15,
			   -ipady => 5, -sticky => 'ew' );
	    
	    $button_ok->configure( -command => sub {
		
		my ( $string ) = $text->get( '1.0', 'end' );
		chomp( $string );
		
		if ( defined( $string ) and length( $string )) {
		    if ( $self->{SAFE_MODE} ) {
			$self->SafeText( $string );
		    }
		    $self->Text( $string );
		    $self->{LAST_INDEX} = $self->index( 'insert' );
		    $button_cancel->invoke();
		}
	    });
	    $button_next->gridForget();
	    $button_all->gridForget();
	}
	
	elsif ( $action eq "Write Tag" ) {
	    
	    $label2 = "Tag Name: ";
	    $frame2->grid( -row => 0, -column => 0, -ipadx => 10,
			   -ipady => 5, -sticky => 'ew' ); 
	    $frame5->grid( -row => 1, -column => 0, -ipadx => 10,
			   -ipady => 5, -sticky => 'ew' );
	    $frame6->grid( -row => 2, -column => 0, -ipadx=>10,
			   -ipady => 5, -sticky => 'ew' );
	    
	    $button_ok->configure( -command => sub {
		
		if ( defined($pattern2) and length($pattern2)) {
		    
		    @atts = ( $attribute1, $value1, $attribute2, $value2,
			      $attribute3, $value3 );
		    
		    if ( $self->{SAFE_MODE} ) {
			
			if ( $radio eq '1' ) {
			    $self->SafeStartTag( $pattern2, @atts );
			}
			elsif ( $radio eq '2' ) {
			    $self->SafeEmptyTag( $pattern2, @atts );
			}
			else {
			    $self->SafeEndTag( $pattern2 );
			}
		    }
		    
		    if ( $radio eq '1' ) {
			$self->StartTag( $pattern2, @atts );
		    }
		    elsif ( $radio eq '2' ) {
			$self->EmptyTag( $pattern2, @atts );
		    }
		    else {
			$self->EndTag( $pattern2 );
		    }
		    $button_cancel->invoke();
		}
	    });
	    $button_next->gridForget();
	    $button_all->gridForget();
	}
	
	else {
	    
	    $label1 = "Find What: ";
	    $label2 = "Tag Name: ";
	    
	    $frame1->grid( -row => 0, -column => 0, -ipadx => 10,
			   -ipady => 5, -sticky => 'ew' );
	    $frame2->grid( -row => 1, -column => 0, -ipadx => 10,
			   -ipady => 5, -sticky => 'ew' );
	    $frame6->grid( -row => 2, -column => 0, -ipadx=>10,
			   -ipady => 5, -sticky => 'ew' );
	    $frame7->grid( -row => 3,-column => 0, -ipadx => 10,
			   -ipady => 5, -sticky => 'w' );
	    
	    $button_ok->configure( -command => sub {
		
		if ( defined( $pattern2 ) and length( $pattern2 )) {
		    
		    @atts = ( $attribute1, $value1, $attribute2, $value2,
			      $attribute3, $value3 );
		    $self->markupSelections( $pattern2, @atts );
		    $button_next->invoke();
		}
	    });
	    $button_next->configure( -command => sub {
		
		$direction = $checked1? '-backward': '-forward';
		$mode      = $checked2? '-regexp': '-exact';
		$case      = $checked3? '-nocase': '-case';
		
		if ( defined( $pattern1 ) and length( $pattern1 )) {
		    $self->FindNext( $direction, $mode, $case, $pattern1 );
		}
	    });
	    $button_all->configure( -command => sub {
		
		if ( defined( $pattern2 ) and length( $pattern2 )) {
		    
		    $mode      = $checked2? '-regexp': '-exact';
		    $case      = $checked3? '-nocase': '-case';
		    @atts = ( $attribute1, $value1, $attribute2, $value2,
			      $attribute3, $value3 );
		    
		    $self->FindAndReplaceAll( \&Tk::TextTools::FindNext,
					\&markupSelections, $mode, $case,
					      $pattern1, $pattern2, @atts );
		}
	    });
	}
    
	$top1->resizable( 0, 0 );
	return( $self->tellStatusbar( "Assisted session opened." ));
    }
    return( 0 );
}


sub showIndex {
    
    my ( $self ) = @_;

    if (! Exists( $top2 )) {
	$self->tellStatusbar( "Preparing the index..." );

	$BTREE = tie %b, "DB_File", undef, O_RDWR|O_CREAT, 0640, $DB_BTREE
	    or die "Cannot tie the database: $!\n";
	$b{locked} = 1;
	
	$top2   = $self->Toplevel();
	my ( $index ) = $top2->Scrolled( 'Listbox', -scrollbars => 'osoe',
					 -background => "#ffefffffe0c3",
					 -selectforeground => 'white',
					 -selectbackground => "#000070669645",
					 -selectmode       => 'extended',
					 -takefocus        => 1,
					 -width            => 50 );
	my ( $menu ) = $top2->Menu();
	my ( $button ) = $menu->Menubutton(

	      -label     => 'File', -tearoff => 0, -underline => 0,
	      -menuitems => [['command' => 'List ~Words',
	      -command   =>  sub {
		  if (! $b{locked} ) {
		      $self->BackTrace( "Could not lock the database. $!\n" );
		  }   
		  $self->fillIndex( $index,'words' );
		  $b{locked} = 1;
	      }],
	      ['command' => 'Tag ~Names',
	       -command   => sub {
		   if (! $b{locked} ) {
		       $self->BackTrace( "Could not lock the database. $!\n" );
		   }   
		   $self->fillIndex( $index,'names' );
		   $b{locked} = 1;
	       }],
	      '-', 
	      ['command' => 'Tag ~Attributes',
	       -command   => sub {
		   if (! $b{locked} ) {
		       $self->BackTrace( "Could not lock the database. $!\n" );
		   }   
		   $self->fillIndex( $index,'attributes' );
		   $b{locked} = 1; 
	      }],
	      ['command' => 'Tag ~Values',
	       -command   => sub {
		   if (! $b{locked} ) {
		       $self->BackTrace( "Could not lock the database. $!\n" );
		   }   
		   $self->fillIndex( $index,'values' );
		   $b{locked} = 1;
	      }],
	      ['command' => 'List ~Comments',
	       -command   => sub {
		   if (! $b{locked} ) {
		       $self->BackTrace( "Could not lock the database. $!\n" );
		   }  
		   $self->fillIndex( $index, 'comments' );
		   $b{locked} = 1;
	      }],
	      '_',
	      ['command' => '~Exit',
	      -command  => sub {

		  if (! $b{locked} ) {
		      $self->BackTrace( "Could not unlock the database. $!\n");
		  }  

		  $index->delete( 0, 'end' );
		  $top2->destroy();

		  undef( $BTREE );
		  untie( %b );
		  
	      }]]);
    
	$top2->configure( -title => "Index", -menu => $menu );
	$button->menu->configure( -postcommand => sub {

	    if ( $self->FileName && $self->FileName =~ /\.xml$/ ) {
		
		$button->menu->entryconfigure( 3, -state => 'normal' );
		$button->menu->entryconfigure( 4, -state => 'normal' );
		$button->menu->entryconfigure( 5, -state => 'normal' );
	    }
	    else {

		$button->menu->entryconfigure( 3, -state => 'disabled' );
		$button->menu->entryconfigure( 4, -state => 'disabled' );
		$button->menu->entryconfigure( 5, -state => 'disabled' );
	    }
	});

	$index->pack( -fill=>'both', -expand => 1 );
	$top2->resizable( 0, 0 );

	return( $self->tellStatusbar( "The index is ready!" ));
    }
    return( 0 )
}


sub fillIndex {
    
    my ( $self, $index, $items ) = @_;
    my ( $status, $key, $value );

    $self->tellStatusbar( "Making index of $items ..." );

    $self->MainWindow->Busy;
    %b = ();

    my ( $parser ) = new XML::Parser();
    if ( $self->FileName && $self->FileName =~ /\.xml$/ ) {
	
	if ( $items eq 'words' ) {
	    $parser->setHandlers( Char => sub {
		
		my ( $expat, $data ) = @_;
		$data = Tk::XMLViewer::convertFromUnicode( $data );
		while ( $data =~ s/(\w+)// ) {
		    $b{$1} = 1;
		}
	    });
	}
	elsif ( $items eq 'names' ) {
	    
	    $parser->setHandlers( Start => sub {
		
		my ( $expat, $name, @atts ) = @_;
		$name = Tk::XMLViewer::convertFromUnicode( $name );
		$b{$name} = 1;
	    });
	}
	elsif ( $items eq 'attributes' ) {
	
	    $parser->setHandlers( Start => sub {
		
		my ( $expat, $name, @atts ) = @_;
		my ( $i );
		
		for ( $i = 0; $i <= $#atts -1; $i+=2 ) {
		    if ( defined( $atts[$i] ) and length( $atts[$i] )) {
		    $atts[$i]= Tk::XMLViewer::convertFromUnicode($atts[$i]);
		    $b{$atts[$i]} = 1;
		    }
		}
	    });
	}
	elsif ( $items eq 'values' ) {
	
	    $parser->setHandlers( Start => sub {
		
		my ( $expat, $name, @atts ) = @_;
		my ( $i );
		
		for ( $i = 1; $i <= $#atts; $i+=2 ) {
		    if ( defined( $atts[$i] ) and length( $atts[$i] )) {
		    $atts[$i]= Tk::XMLViewer::convertFromUnicode($atts[$i]);
		    $b{$atts[$i]} = 1;
		    }
		}
	   });
	}
	else {

	    $parser->setHandlers( Comment => sub {
		
		my ( $expat, $data ) = @_;
		$data = Tk::XMLViewer::convertFromUnicode( $data );
		$b{$data} = 1;
	    });
	}

	eval { $parser->parse( $self->get( '1.0', 'end' )); };
	if ( $@ ) {
	    %b = ( "locked" => 1 );
	    return( $self->BackTrace( "Parse error!" ));
	}
    }

    else {
	
	my ( $progress, $count, $line, $end, $i );
	
	$end   = $self->index( 'end' );
	$end   =~ s/\.\d+$//;
	
	for ( $i = 1; $i < $end; $i++ ) {
	    if (( $count++%100 ) == 0 ) {
		
		$progress = $self->TextUndoFileProgress(
		"Making index"  => $self->FileName(), $count, $i, $end-1 );
		
	    }

	    $line = $self->get( "$i.0", "$i.end" );
	    if ( $items eq 'names' ) {
		while ( $line =~ s/<(\w+)[^>]*>// ) {
		    $b{$1} = 1;
		}
	    }
	    elsif ( $items eq 'words' ) {
		$line =~ s/<[^>]+>//g;
		while ( $line =~ s/(\w+)// ) {
		    $b{$1} = 1;
		}
	    }
	}
	
	if ( defined ( $progress )) {
	    $progress->withdraw();
	}
    }
    
    $index->delete( 0, 'end' );
    $key = $value = 0;

    for ( $status = $BTREE->seq( $key, $value, R_FIRST );
	  $status == 0 ; $status = $BTREE->seq( $key, $value, R_NEXT)) {
	$index->insert( 'end', $key );
    }

    $self->MainWindow->Unbusy;
    return( $self->tellStatusbar( "Made index of $items." ));
}


sub showPreferences {
    
    my ( $self ) = @_;
    if (! Exists( $top3 )) {

	$self->tellStatusbar( "Showing preferences..." );
	$top3 = $self->Toplevel( -title => "Preferences" );
    
	my ( $frame1 ) = $top3->Frame( -relief=> 'groove', -borderwidth=> 2 )->grid
	    ( -row=> 0,-column=> 0, -ipadx=>10, -ipady=> 10,-sticky=> 'nsew' );
	my ( $button1 ) = $frame1->Menubutton( -text    => 'Background Color',
					       -relief  => 'raised', -width => 45,
					       -tearoff   => 0,
					       -underline => 0 )->grid;

	$button1->command( -label => "~white", -background => 'white',
			   -command => sub { $self->{BACKGROUND}='white';});
	$button1->command( -label => "~gainsboro", -background => "#dcecdcecdcec",
			   -command=> sub { $self->{BACKGROUND}="#dcecdcecdcec";});
	$button1->command( -label => "~floral white", -background=>"#fa9ef5aeebd5",
			   -command => sub{$self->{BACKGROUND}="#fa9ef5aeebd5"; });
	$button1->command( -label => "~azure", -background => "#cd7bda5dda5d",
			   -command =>sub{$self->{BACKGROUND}="#cd7bda5dda5d";});
	$button1->command( -label => "~cornsilk", -background =>"#fffff8e9dcec",
			   -command =>sub{$self->{BACKGROUND}="#fffff8e9dcec";});
	$button1->command( -label => "~honeydew", -background =>"#c363cf9cc35d",
			   -command =>sub{$self->{BACKGROUND}="#c363cf9cc35d";});
	
	my ( $button2 ) = $frame1->Menubutton( -text    => 'Font Definition',
					       -relief  => 'raised', -width => 45,
					       -tearoff => 0, -underline => 0 )->grid;
	
	my ( $cascade1 ) = $button2->cascade( -label => "~Family", -tearoff => 0 );
	foreach ( qw/times helvetica arial courier/ ) {
	    $cascade1->radiobutton( -label => "~$_", -value => $_,
				    -variable => \$self->{FONT_FAMILY} );
	}
	my ( $cascade2 ) = $button2->cascade( -label => "~Point Size",
					      -tearoff=> 0 ) ;
	foreach ( qw/10 12 14 16 18 24 32/ ) {
	    $cascade2->radiobutton( -label => "$_ point", -value => $_,
				    -variable => \$self->{FONT_SIZE} );
	}
	my ( $cascade3 ) = $button2->cascade( -label => "~Weight", -tearoff=> 0 );
	foreach ( qw/normal bold/ ) {
	    $cascade3->radiobutton( -label => "~$_", -value => $_,
				    -variable => \$self->{FONT_WEIGHT} );
	}
	my ( $cascade4 ) = $button2->cascade( -label => "~Slant", -tearoff=> 0 );
	foreach ( qw/roman italic/ ) {
	    $cascade4->radiobutton( -label => "~$_", -value => $_,
				    -variable => \$self->{FONT_SLANT} );
	}
	
	my ( $button3 ) = $frame1->Menubutton( -text => 'Indent String',
					       -relief => 'raised', -width => 45,
					       -tearoff   => 0,
					       -underline => 0 )->grid;
	
	$button3->radiobutton( -label => "~Double Space", -value => "  ",
			       -variable => \$self->{INDENT_STRING} );
	$button3->radiobutton( -label => "~Tabulation", -value => "\t",
			       -variable => \$self->{INDENT_STRING} );
	
	
	my ( $frame2 ) = $top3->Frame()->grid
	    ( -row => 1,-column => 0, -ipadx => 10, -sticky => 'ew' );
	
	$frame2->Checkbutton( -text => "Safe Mode",
			      -variable => \$self->{SAFE_MODE} )->grid
				  ( -row => 0, -column => 0, -sticky => 'w',
				    -ipadx => 5 );
	$frame2->Checkbutton( -text => "Data Mode",
			      -variable => \$self->{DATA_MODE} )->grid
				  ( -row => 0, -column => 1, -sticky=>'w',
				    -ipadx => 5 );
	$frame2->Checkbutton( -text => "Data Indent",
			      -variable => \$self->{DATA_INDENT} )->grid
				  ( -row =>0, -column=>2, -sticky=>'w',
				    -ipadx => 5 );
	
	my ( $frame3 ) = $top3->Frame( -relief=> 'groove', -borderwidth=> 2 )->grid
	    ( -row => 2,-column => 0, -ipadx => 10, -ipady => 10,-sticky => 'ew' );
	my ( $button_configure ) = $frame3->Button( -text => 'Configure',
						    -command=>sub {
							$self->loadEditPreferences();})->grid
							    ( -row => 0, -column => 0,
							      -sticky => 'nsew', -padx => 30 );
	my ( $button_save ) = $frame3->Button( -text => 'Save', -command => sub {
	    
	    if ( $self->{HOME} && open( CONF, ">".$self->{HOME}."/.edxmlrc" )){}
	    elsif ( open( CONF, ">".$Bin.".edxmlrc" )) {}
	    else { die "Cannot write the configuration file! $!\n" ; }
	    
	    print CONF <<EOM;
SAFE_MODE=$self->{SAFE_MODE}
DATA_MODE=$self->{DATA_MODE}
DATA_INDENT=$self->{DATA_INDENT}
INDENT_STRING=$self->{INDENT_STRING}
FONT_FAMILY=$self->{FONT_FAMILY}
FONT_SIZE=$self->{FONT_SIZE}
FONT_WEIGHT=$self->{FONT_WEIGHT}
FONT_SLANT=$self->{FONT_SLANT}
BACKGROUND=$self->{BACKGROUND}
EOM

    close( CONF );
	})->grid( -row => 0, -column => 1,-sticky => 'nsew', -padx => 15 );
	my ( $button_dismiss ) = $frame3->Button( -text => 'Dismiss',
						  -command => sub {
						      $self->tellStatusbar( "Configuration session closed." );
						      return( $top3->destroy() );
						  })->grid( -row=>0,-column=>2,-sticky=>'nsew', -padx => 30 );
	
	$top3->resizable( 0, 0 );
	return( $self->tellStatusbar( "Menu 'Preferences' displayed." ));  
    }
}


sub loadEditPreferences {

    my ( $self ) = @_;
    
    $self->configure( -background => $self->{BACKGROUND},
		      -font       => $self->fontCreate

	( -family => $self->{FONT_FAMILY},
	  -size   => $self->{FONT_SIZE},
	  -weight => $self->{FONT_WEIGHT},
	  -slant  => $self->{FONT_SLANT}));

    return( $self );
}


sub getSafeMode {
    
    my ( $self ) = @_;
    return( $self->{SAFE_MODE} );
}


sub getLastIndex {
    
    my ( $self ) = @_;
    return( $self->{LAST_INDEX} );
}


sub getXMLDecl {

    my ( $self ) = @_;
    return( ${$self->{SEEN}}{XMLDECL} );
}


sub getDoctype {

    my ( $self ) = @_;
    return( ${$self->{SEEN}}{DOCTYPE} );
}


1;

__END__


=head1 NAME
    
MarkupTools - perl extensions library for "edxml"
    
=head1 SYNOPSIS
    
use MarkupTools;

$buffer= $toplevel->Scrolled( 'MarkupTools' );
$buffer->setHomeEnvironement( $ENV{HOME}? $ENV{HOME}: 0 );
$buffer->loadEditPreferences();

$menu  = $buffer->menu();

=head1 DESCRIPTION

This library provides an interface to the standard perl
routines that manipulate XML objects from XML::Parser.pm.

=head1 AUTHOR
    
Calin MOSUT, "Babes-Bolyai" University
Faculty of Letters, A.M.L. Departement
Bulevard Horea, 31, 3400 Cluj-Napoca
Romania

Phone : (40/)64 432 540
Email : cmosut@personal.ro

=cut
